"""
Simulation results object and function to initialize it from simulation pipeline output.
"""

__all__ = ["SimResults", "make_results"]

from .make_results import make_results
from .sim_results import SimResults
